C********************************************************************
C     *** FACTOR ... CHOLESKY FACATORIZATION OF  A*THETA*Atransp ***
C********************************************************************
C
      SUBROUTINE FACTOR(MAXM,MAXN,MAXNZA,MAXNZL,M,
     X IWORK,RWORK,IMAP,RMAP,LIWORK,LRWORK,
     X ITEMP1,RTEMP1,
     X HEADER,LINKFD,LINKBK,
     X THETA,STAVAR,
     X LCOEFF,LDIAG,LDSQRT,LCLPTS,LRWNBS,MKSQRT,RTCODE,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXM,MAXN,MAXNZA,MAXNZL,M,LIWORK,LRWORK
      INTEGER*4 MKSQRT,RTCODE,IOLOG
      INTEGER*4 ITEMP1(MAXN)
      DOUBLE PRECISION RTEMP1(MAXM)
      INTEGER*2 HEADER(MAXM+1),LINKFD(MAXM+1),LINKBK(MAXM+1)
      DOUBLE PRECISION THETA(MAXN)
      INTEGER*2 STAVAR(MAXN)
C
C *** DATA STRUCTURES FOR CHOLESKY FACTOR
      DOUBLE PRECISION LCOEFF(*)
      DOUBLE PRECISION LDIAG(MAXM),LDSQRT(MAXM)
      INTEGER*4 LCLPTS(MAXM+1)
      INTEGER*2 LRWNBS(MAXNZL)
C
C *** HIDDEN DATA STRUCTURES
      INTEGER*4 IWORK(LIWORK),IMAP(*),RMAP(*)
      DOUBLE PRECISION RWORK(LRWORK)
C
C
C
C *** LOCAL VARIABLES
C
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     MAXN    Maximum column dimension of the LP constraint matrix.
C     MAXNZA  Maximum number of nonzeros of the LP constraint matrix.
C     MAXNZL  Maximum number of nonzeros of the Cholesky factor.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     ITEMP1  Integer work array.
C     RTEMP1  Double precision work array.
C     HEADER  Header of the doubly linked lists.
C     LINKFD  Forward linked lists.
C     LINKBK  Backward linked lists.
C     THETA   Diagonal weight matrix.
C     STAVAR  Array of variable status:
C             0  STANDARD (nonnegative) variable i.e.: 0 <= x <= +inf;
C             1  UPPER bounded variable i.e.: 0 <= x <= u;
C             2  LOWER bounded variable i.e.: l <= x <= +inf;
C             3  both LOWER and UPPER bounded variable i.e.: l <= x <= u;
C             4  MINUS INFINITY type variable i.e.: -inf <= x <= u;
C             5  PLUS INFINITY type variable i.e.: l <= x <= +inf;
C             6  FIXED variable i.e.: x = l = u;
C            -k  free variable. As the free variable is split into two
C                variables,  k is the number of its brother. Observe,
C                that k-th variable will also have negative status
C                that indicate the position of the original variable.
C     MKSQRT  Parameter indicating if square roots of LDIAG are to be
C             computed:
C             0  no square roots necessary;
C             1  compute square roots of diagonal matrix.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     RTCODE  Indicator if there are errors in the Cholesky factors:
C             0 if Cholesky factor is sufficiently stable;
C             1 if Cholesky factorization failed.
C     LCOEFF  Off-diagonal nonzero coefficients of Cholesky matrix.
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C     LDIAG   Diagonal elements of Cholesky factor.
C     LDSQRT  Square roots of the diagonal elements of Cholesky factor.
C
C
C
C *** HIDDEN DATA STRUCTURES DESCRIPTION
C     RWORK   Real work array containing almost all real
C             LP problem data.
C     IWORK   Integer work array containing almost all integer
C             LP problem data.
C     RMAP    Map of RWORK.
C     IMAP    Map of IWORK.
C
C     Map of IWORK array:
C     IMAP(1)   Points to CLPNTS array.
C     IMAP(2)   Points to RWNMBS array.
C     IMAP(3)   Points to RWHEAD array.
C     IMAP(4)   Points to RWLINK array.
C     IMAP(5)   Points to CLNMBS array.
C     IMAP(6)   Points to LENCOL array.
C     IMAP(7)   Points to the first empty cell of IWORK array.
C
C     Map of RWORK array:
C     RMAP(1)   Points to ACOEFF array.
C     RMAP(2)   Points to COBJ array.
C     RMAP(3)   Points to RHS array.
C     RMAP(4)   Points to the first empty cell of RWORK array.
C
C
C
C
C
C *** SUBROUTINES CALLED:
C     LDAAT,NUMFCT
C
C
C *** PURPOSE:
C     This routine computes Cholesky decomposition
C       L*D*Ltransp  of  A*THETA*Atransp.
C
C     It does this in the following two steps:
C     (i)    construction of  A*THETA*Atransp matrix and packing
C            it in a data structures for the Cholesky factor
C            (fill-in positions are zeroed).
C     (ii)   computing the Cholesky decomposition.
C
C
C
C *** NOTES:
C     This routine is an interface between  IPMLO library for
C     linear optimization with interior point methods and the
C     library of routines for handling the Cholesky decomposition
C     of a sparse positive definite systems.
C
C
C *** REFERENCES:
C     Gondzio J. (1993). Implementing Cholesky factorization
C        for interior point methods of linear programming,
C        Optimization 27, pp. 121-140.
C     Gondzio J., Tachat D. (1994). The design and application of 
C        IPMLO - a FORTRAN library for linear optimization with 
C        interior point methods, RAIRO Recherche Operationnelle 28, 
C        No 1, pp. 37-56.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  May 19, 1991
C     Last modified: September 25, 1995
C
C
C
C *** BODY OF (FACTOR) ***
C
C
C     Load the  A*THETA*Atransp matrix into the data structures
C     for Cholesky factor  L (zero all fill-in positions).
C
      CALL LDAAT(RWORK(RMAP(1)),IWORK(IMAP(1)),IWORK(IMAP(2)),
     X IWORK(IMAP(3)),IWORK(IMAP(4)),IWORK(IMAP(5)),IWORK(IMAP(6)),
     X THETA,STAVAR,MAXNZL,MAXM,MAXN,MAXNZA,M,
     X LCOEFF,LCLPTS,LRWNBS,LDIAG,RTEMP1,IOLOG)
C
C
C     Decompose the  A*THETA*Atransp matrix.
C
      CALL NUMFCT(LCOEFF,LCLPTS,LRWNBS,LDIAG,LDSQRT,
     X MAXNZL,MAXM,M,MKSQRT,RTCODE,
     X HEADER,LINKFD,LINKBK,ITEMP1,RTEMP1,IOLOG)
C
C
      RETURN
C
C *** LAST CARD OF (FACTOR) ***
      END
