C***********************************************************************
C     * SCHUR2 ... SOLVE EQUATION WITH (A,F)*(THETA,THF)*(A,F)transp *
C     *            TWO COLUMNS ARE BORDERED TO THE LP CONSTRAINT MTX * 
C***********************************************************************
C
      SUBROUTINE SCHUR2(MAXNZL,MAXM,M,COLUMN,RHS,
     X LCOEFF,LCLPTS,LRWNBS,LDIAG,LDSQRT,
     X FCLMN1,FCLMN2,THF1,THF2,
     X WCLMN1,WCLMN2,RTEMP1,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXNZL,MAXM,M,IOLOG
      DOUBLE PRECISION LCOEFF(*),LDIAG(MAXM),LDSQRT(MAXM)
      DOUBLE PRECISION COLUMN(MAXM),RHS(MAXM)
      DOUBLE PRECISION FCLMN1(MAXM),FCLMN2(MAXM),THF1,THF2
      DOUBLE PRECISION WCLMN1(MAXM),WCLMN2(MAXM),RTEMP1(MAXM)
      INTEGER*4 LCLPTS(MAXM+1)
C
C *** The following array can be half-length integer.
      INTEGER*2 LRWNBS(MAXNZL)
C
C
C *** LOCAL VARIABLES
      DOUBLE PRECISION HCOL(2),ZCOL(2),DP1,DP2
      DOUBLE PRECISION SCHR11,SCHR21,SCHR22,SD11,SD22,SL21
      INTEGER*4 IROW
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:  Cholesky factor of  A*THETA*Atransp.
C     MAXNZL  Maximum number of nonzeros of the Cholesky factor.
C     MAXM    Maximum row dimension of the LP constraint matrix.
C     M       Number of rows of the LP constraint matrix
C             (and the dimension of  A*Atransp).
C     LCOEFF  Off-diagonal nonzero coefficients of Cholesky matrix.
C     LCLPTS  Pointers to columns of the Cholesky factor.
C     LRWNBS  Row numbers of nonzeros in columns of matrix  L.
C     LDIAG   Diagonal elements of Cholesky factor.
C     LDSQRT  Square root of the diagonal matrix of the Cholesky
C             decomposition.
C     FCLMN1  First  column of matrix  F (supposed to be dense).
C     FCLMN2  Second column of matrix  F (supposed to be dense).
C     THF1    First  element of matrix  THETAF.
C     THF2    Second element of matrix  THETAF.
C     RHS     Right-hand-side of the equation.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     COLUMN  Solution of the equation
C             ( (A,F)*(THETA,THF)*(A,F)transp ) * X = RHS.
C
C     WORK ARRAYS:
C     WCLMN1  First  column of matrix  W.
C     WCLMN2  Second column of matrix  W.
C     SCHRij  Element  S(i,j) of the  2x2 symmetric Schur complement.
C     RTEMP1  Temporary work array.
C
C
C
C *** SUBROUTINES CALLED:
C     DSQRT,DDOT,SOLVL,SOLAAT
C
C
C *** PURPOSE:
C     This routine solves equation with  (A,F)*(THETA,THF)*(A,F)transp.
C     It handles columns of  F implicitly to avoid their degrading
C     influence on the sparsity of Cholesky factor  L.
C     It uses the Cholesky decomposition  L*D*Ltransp of  A*THETA*Atransp
C     (the decomposition must be computed before calling this routine).
C
C
C *** NOTES:
C     1. The contents of  RHS array is destroyed by this routine.
C
C     2. This routine is compatible with Gondzio's implementation
C        of Cholesky decomposition.
C
C     3. This routine performs the following sequence of calculations:
C     WCLMN1  Column  w1:   solve eqn  (L*D**0.5) * w1 = f1.
C     WCLMN2  Column  w2:   solve eqn  (L*D**0.5) * w2 = f2.
C     SCHRij  2x2  Schur:   compute    I + (THF**0.5)*(Wtransp*W)*(THF**0.5)
C     COLUMN  Column   g:   solve eqn  (L*D**0.5) * g = d.
C     HCOL    Column   h:   compute    h = THF**0.5 * Wtransp * g.
C     ZCOL    Column   z:   solve eqn  S * z = h.
C     RHS     Column   t:   compute    t = d - F * (THF**0.5) * z.
C     COLUMN  Column   x:   solve eqn  (L*D*Ltransp) * x = t.
C
C
C *** REFERENCES:
C     Cottle R.W. (1974). Manifestations of the Schur complement,
C        Linear Algebra and its Applications, vol 8, pp. 189-211.
C     Choi I.C., Monma C.L., Shanno D.F. (1990). Further development
C        of a primal-dual interior point method, ORSA Journal
C        on Computing, vol 2, pp. 304-311.
C     Gondzio J. (1993). Implementing Cholesky factorization for 
C        interior point methods of linear programming, Optimization 
C        27, pp. 121-140.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  September 18, 1991
C     Last modified: May 15, 1995
C
C
C
C *** BODY OF (SCHUR2) ***
C
C
C
C
C     Copy  FCLMN1 array into  RTEMP1 array.
C     Copy  FCLMN2 array into  COLUMN array.
      DO 100 IROW=1,M
         RTEMP1(IROW)=FCLMN1(IROW)
         COLUMN(IROW)=FCLMN2(IROW)
  100 CONTINUE
C
C
C
C
C     Solve the equation  (L*D**0.5) * w1 = f1.
C     Solve the equation  (L*D**0.5) * w2 = f2.
C     Next, save  RHS in  RTEMP1 array.
C
      CALL SOLVL(LCOEFF,LCLPTS,LRWNBS,
     X MAXNZL,MAXM,M,WCLMN1,RTEMP1,IOLOG)
      CALL SOLVL(LCOEFF,LCLPTS,LRWNBS,
     X MAXNZL,MAXM,M,WCLMN2,COLUMN,IOLOG)
C
      DO 200 IROW=1,M
         WCLMN1(IROW)=WCLMN1(IROW)/LDSQRT(IROW)
         WCLMN2(IROW)=WCLMN2(IROW)/LDSQRT(IROW)
         RTEMP1(IROW)=RHS(IROW)
  200 CONTINUE
C
C
C
C
C     Build up the  2x2 Schur complement.
C     I + (THF**0.5)*(Wtransp*W)*(THF**0.5)
C
      CALL DDOT(WCLMN1,WCLMN1,M,SCHR11)
      CALL DDOT(WCLMN1,WCLMN2,M,SCHR21)
      CALL DDOT(WCLMN2,WCLMN2,M,SCHR22)
C
      SCHR11=SCHR11*THF1+1.0
      SCHR21=SCHR21*DSQRT(THF1*THF2)
      SCHR22=SCHR22*THF2+1.0
C
C
C
C
C     Solve the equation  (L*D**0.5) * g = d.
C
      CALL SOLVL(LCOEFF,LCLPTS,LRWNBS,
     X MAXNZL,MAXM,M,COLUMN,RTEMP1,IOLOG)
C
      DO 300 IROW=1,M
         COLUMN(IROW)=COLUMN(IROW)/LDSQRT(IROW)
  300 CONTINUE
C
C
C
C
C     Compute    h = THF**0.5 * Wtransp * g.
C
      CALL DDOT(WCLMN1,COLUMN,M,HCOL(1))
      CALL DDOT(WCLMN2,COLUMN,M,HCOL(2))
      HCOL(1)=HCOL(1)*DSQRT(THF1)
      HCOL(2)=HCOL(2)*DSQRT(THF2)
C
C
C
C
C     Solve equation with a  2x2 Schur complement   S * z = h.
C
      SD11=SCHR11
      SL21=SCHR21/SD11
      SD22=SCHR22-SL21*SCHR21
C
      ZCOL(1)=HCOL(1)
      ZCOL(2)=HCOL(2)-SL21*ZCOL(1)
C
      ZCOL(1)=ZCOL(1)/SD11
      ZCOL(2)=ZCOL(2)/SD22
C
C     ZCOL(2)=ZCOL(2)
      ZCOL(1)=ZCOL(1)-SL21*ZCOL(2)
C
C
C
C
C
C     Compute    t = d - F * (THF**0.5) * z.
C
      DP1=ZCOL(1)*DSQRT(THF1)
      DP2=ZCOL(2)*DSQRT(THF2)
C
      DO 400 IROW=1,M
         RHS(IROW)=RHS(IROW)-FCLMN1(IROW)*DP1-FCLMN2(IROW)*DP2
  400 CONTINUE
C
C
C
C
C     Solve the equation  (L*D*Ltransp) * x = t.
C
      CALL SOLAAT(LCOEFF,LCLPTS,LRWNBS,LDIAG,
     X MAXNZL,MAXM,M,COLUMN,RHS,IOLOG)
C
C
C
C
      RETURN
C
C
C *** LAST CARD OF (SCHUR2) ***
      END
